#include "config.h"
#include "system.h"
#include "tree.h"
#include "ggc.h"
#include "flags.h"
#include "toplev.h"
#include "convert.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* I don't know if they work correctly... :<, #undef to disable */
#define LABELS

/* internal compiler error */
#define ice() abort()

typedef enum {
#define KW(t) kw_ ## t ,
#include "kw.def"
#undef KW
	kw__error
} keyword_t;

typedef enum {
	storage_error,
	storage_local,
	storage_public,
	storage_extern,
} storage_t;

/* for ALLOWED argument for get*arg*() functions */
enum {
	arg_int		= 0x0001,
	arg_uint	= 0x0002,
	arg_float	= 0x0004,
	arg_string	= 0x0008,
	arg_id		= 0x0010,
	arg_kw		= 0x0020,
};

struct arg {
	int sel;
	union {
		struct {
			HOST_WIDE_INT low;
			HOST_WIDE_INT high;
		} val_int;
		struct {
			unsigned HOST_WIDE_INT low;
			unsigned HOST_WIDE_INT high;
		} val_uint;
		/* unfortunetly, there is no better way, then ASCII */
		/* representation. */
		const char *val_float;
		struct {
			const char *data;
			int len;
		} val_string;
		const char *val_id;
		keyword_t val_kw;
	} u;
};

/* should be abstract */
struct term_state {
	keyword_t kw;
	int arg_no, next_sep;
};

/* should be abstract */
struct list_state {
	keyword_t kw;
	int arg_no, lev, list;
};

extern tree bool_type_node;
extern int flag_just_preprocess;
extern int flag_yydebug;

/* glue.c */
void init_types(void);
void fini_types(void);
int lvalue_p(tree ref);

/* go up in scopes, looking for name */
tree lookup_id(const char *text, enum tree_code type);
/* look for TYPE_DECL, declaring type of TREE_CODE TYPE */
tree lookup_type(tree id, enum tree_code type);
tree get_vars(void);

/* expr.c */
tree compile_unary(keyword_t op, enum tree_code code, tree *type);
tree compile_bin(keyword_t op, enum tree_code code, tree *type);
tree compile_const(tree *type);
tree compile_compound(tree *type);
void same_types(keyword_t kw, tree t1, tree t2);
tree compile_cond(tree *type);
tree compile_indirect(tree *type);
tree compile_modify(tree *type);
tree compile_nop(tree *type);
tree compile_call(tree *type, int side);
tree compile_result(tree *type);
tree compile_ref(tree *type);
tree compile_var(tree *type, int static_p);
tree compile_convert(tree *type);
tree compile_float(tree *type);
tree compile_fix(tree *type);
tree compile(keyword_t kw, tree *type);
tree compile_no_kw(tree *type);
tree compile_component(tree *type);

/* lex.c */
/* low-level token fetching routines */
int get_starter(void);
int get_separator(void);
struct arg *get_arg(int allowed, struct arg *ret);
void kpp_init(void);
void kpp_define_opt(const char *p);

/* lookups */
const char *keyword_name(keyword_t kw);
const char *arg_name(int arg);

/* simple term handling */
void init_term(struct term_state *ctx, keyword_t kw);
void fini_term(struct term_state *ctx);
void arg0_term(keyword_t kw);
void get_term_arg(struct term_state *ctx, int allowed, struct arg *a);
tree get_id(struct term_state *ctx);
int optional_term_arg_present(struct term_state *ctx);
keyword_t get_keyword(struct term_state *ctx);
void expect_keyword(struct term_state *ctx, keyword_t exp);

/* list term handling */
void init_list(struct list_state *ctx, keyword_t kw);
int get_list_arg(struct list_state *ctx, int allowed, struct arg *a);
keyword_t get_keyword_from_list(struct list_state *ctx);
void expect_keyword_in_list(struct list_state *ctx, keyword_t exp);

/* stmts.c */
tree compile_bind(tree *type);
tree compile_return(tree *type);
tree compile_loop(tree *type);
tree compile_exit_loop(tree *type);
tree compile_block(tree *type);
tree compile_exit(tree *type);
#ifdef LABELS
tree compile_label(tree *type);
tree compile_goto(tree *type);
#endif
void fini_labels(void);
tree compile_wfl(tree *type);
void parse_input_file(void);

/* unit.c */
const char *storage_name(storage_t s);
storage_t lookup_storage(const char *s);
void parse_func(void);
void parse_global(void);
void parse_unit(void);
void parse_structdef(int uniondef);

/* type.c */
int type_keyword(keyword_t kw);
tree parse_type(keyword_t kw);
void parse_var(tree *type, tree *name);

struct lang_identifier
{
	struct tree_identifier ignore;
	const char *system_macro;
	const char *user_macro;
};

/* User/system macros for kpp */
#define IDENTIFIER_SYSTEM_MACRO(NODE)	\
  (((struct lang_identifier *) (NODE))->system_macro)
#define IDENTIFIER_USER_MACRO(NODE)	\
  (((struct lang_identifier *) (NODE))->user_macro)
